//----------------------------------------------------------------------------------
//-- ASCON_AEAD
//-- From VHDL version by Behnaz Rezvani
//   --> see _baseline version
//----------------------------------------------------------------------------------
// This version modified with efforts from Ryan Robucci, Vajesh Mistry, Saurin Shah, Chintan Patel
// refactoring and verification underway, code for reference only, untested
module ASCON128 //begin architecture ASCON128
  (
   input wire            clk,
   input wire            rst,
        //-- Data Input
   input wire [ 31: 0]   bdi, //-- W = 32
     //-- BDI Control
   input wire            bdi_valid,
   output logic          bdi_ready,
   input wire [ 3: 0]    bdi_pad_loc, //NOTE: unused -- W/8 = 4
   input wire [ 3: 0]    bdi_valid_bytes, //-- W/8 = 4
   input wire [ 2: 0]    bdi_size, //-- W/(8+1) = 3
   input wire            bdi_eot,
   input wire            bdi_eoi,
   input wire [ 3: 0]    bdi_type,
   input wire            hash_in,
   input wire            decrypt_in,
        //-- Data Output
   output logic [ 31: 0] bdo, //-- W = 32
        //-- BDO Control
   output logic          bdo_valid,
   input wire            bdo_ready,
   output logic [ 3: 0]  bdo_valid_bytes, //-- W/8 = 4
   output logic          end_of_block,
   //output logic [ 3: 0]  bdo_type, //NOTE: removed
        //-- Tag Verification
   output logic          msg_auth,
   output logic          msg_auth_valid,
   input wire            msg_auth_ready,
        //keg input
   input wire [ 127: 0]  key_tag_reg

    );
//end port list of ASCON128

import SomeFunctions::*;

//-- Architecture
//----------------------------------------------------------------------------------
 //begin architecture Behavioral

    //-- Constants -----------------------------------------------------------------
    //--bdi_type and bdo_type encoding
    localparam              HDR_AD = 4'b0001;
    localparam              HDR_PT = 4'b0100; //changed to HDR_MSG in new VHDL
    localparam              HDR_CT = 4'b0101;
    localparam             HDR_TAG = 4'b1000;
    localparam             HDR_KEY = 4'b1100;
    localparam            HDR_NPUB = 4'b1101;
    localparam        HDR_HASH_MSG = 4'b0111;
    localparam      HDR_HASH_VALUE = 4'b1001;

    localparam              zero32 = 0;
    localparam                  IV = 64'h80400c0600000000; //-- k||r||a||b||0*
    localparam             IV_hash = 64'h00400c0000000100; //-- 0*||r||a||0*||h

    //-- Types ---------------------------------------------------------------------
    typedef enum { idle,                     
                   load_key,                
                   wait_Npub,               
                   load_Npub,                                 
                   Initialization,                  
                   Initialization_pause,                  
                   wait_AD,                 
                   load_AD,                 
                   process_AD,                  
                   process_last_AD,                  
                   wait_data,                  
                   load_data,                  
                   process_data,                 
                   Finalization,                 
                   output_tag,                  
                   wait_tag,                  
                   load_tag,                  
                   load_tag_pause,                  
                   verify_tag,                  
                   process_hash} fsm;       

    //-- Signals -------------------------------------------------------------------
    //-- Permutation signals
    logic               perm_start       ;
    logic               a_rounds         ;
    logic [  63:   0]                  Sr;
    logic [ 127:   0]            Sc0,                  Sc1;
    logic [  63:   0]             perm_Sr;
    logic [ 127:   0]            perm_Sc0;
    logic [ 127:   0]            perm_Sc1;
    logic               perm_done        ;

    //-- Data signals

    logic [  31:   0]               bdo_t;

    //-- Control Signals
    logic               decrypt_rst      ;
    logic               decrypt_set      ;
    logic               decrypt_reg      ;

    logic               hash_reg         ;
    logic               hash_rst         ;
    logic               hash_set         ;

    logic               last_AD_reg      ;
    logic               last_AD_rst      ;
    logic               last_AD_set      ;

    logic               no_AD_reg        ;
    logic               no_AD_rst        ;
    logic               no_AD_set        ;

    logic               partial_AD_reg   ;
    logic               partial_AD_rst   ;
    logic               partial_AD_set   ;

    logic               last_M_reg       ;
    logic               last_M_rst       ;
    logic               last_M_set       ;

    logic               no_M_reg         ;
    logic               no_M_rst         ;
    logic               no_M_set         ;

    logic               partial_M_reg    ;
    logic               partial_M_rst    ;
    logic               partial_M_set    ;

    logic               tag_check_reg    ;
    logic               tag_check_rst    ;
    logic               tag_check_set    ;

    //-- Counter signals
    logic               ctr_words_rst    ;
    logic               ctr_words_inc    ;
    logic [   1:   0]           ctr_words;
    logic               ctr_hash_rst     ;
    logic               ctr_hash_inc     ;
    logic [   2:   0]            ctr_hash;

     //-- State machine signals
    fsm state,                  next_state;

    //-- new signals for ROT
    logic               load_en_sc1      ;

//------------------------------------------------------------------------------
/*begin*/

    Permutation P // library work //-- The SPN permutation (pc . ps. pl)
        (
            .clk( clk),
            .rst( rst),
            .start( perm_start),
            .a_rounds( a_rounds),
            .Sr( Sr),
            .Sc0( Sc0),
            .Sc1( Sc1),
            .pSr( perm_Sr),
            .pSc0( perm_Sc0),
            .pSc1( perm_Sc1),
            .done( perm_done)
        );


    assign bdo = bdo_t;
    //--bdo_type <= bdi_type xor "0001" when (bdi_type = HDR_MSG or bdi_type = HDR_CT) else HDR_TAG; -- HDR_CT = HDR_MSG xor "0001"
    assign bdo_valid_bytes = (bdi_type  ==  HDR_PT || bdi_type  ==  HDR_CT) ? bdi_valid_bytes : 4'b1111;

    //---------------------------------------------------------------------------------
    always_ff @ (posedge clk) begin: Sync 
    
       if (rst  ==  1'b1) begin
          state <= idle;
       end else begin
          state <= next_state;
       end
       
       if (ctr_words_rst  ==  1'b1) begin
          ctr_words <= 2'b00;
       end else if (ctr_words_inc  ==  1'b1) begin
          ctr_words <= {ctr_words + 2'b1};
       end
       
       if (ctr_hash_rst  ==  1'b1) begin
          ctr_hash <= 3'd0;
       end else if (ctr_hash_inc  ==  1'b1) begin
          ctr_hash <= {ctr_hash + 3'b1};
       end
       
       if (decrypt_rst  ==  1'b1) begin
          decrypt_reg <= 1'b0;
       end else if (decrypt_set  ==  1'b1) begin
          decrypt_reg <= 1'b1;
       end
       
       if (hash_set  ==  1'b1) begin
          hash_reg <= 1'b1;
       end else if (hash_rst  ==  1'b1) begin
          hash_reg <= 1'b0;
       end
       
       if (last_AD_rst  ==  1'b1) begin
          last_AD_reg <= 1'b0;
       end else if (last_AD_set  ==  1'b1) begin
          last_AD_reg <= 1'b1;
       end
       
       if (no_AD_rst  ==  1'b1) begin
          no_AD_reg <= 1'b0;
       end else if (no_AD_set  ==  1'b1) begin
          no_AD_reg <= 1'b1;
       end
       
       if (partial_AD_rst  ==  1'b1) begin
          partial_AD_reg <= 1'b0;
       end else if (partial_AD_set  ==  1'b1) begin
          partial_AD_reg <= 1'b1;
       end
       
       if (last_M_rst  ==  1'b1) begin
          last_M_reg <= 1'b0;
       end else if (last_M_set  ==  1'b1) begin
          last_M_reg <= 1'b1;
       end
       
       if (no_M_rst  ==  1'b1) begin
          no_M_reg <= 1'b0;
       end else if (no_M_set  ==  1'b1) begin
          no_M_reg <= 1'b1;
       end
       
       if (partial_M_rst  ==  1'b1) begin
          partial_M_reg <= 1'b0;
       end else if (partial_M_set  ==  1'b1) begin
          partial_M_reg <= 1'b1;
       end
       
       if (tag_check_rst  ==  1'b1) begin
          tag_check_reg <= 1'b0;
       end else if (tag_check_set  ==  1'b1) begin
          tag_check_reg <= 1'b1;
       end

    end //process 

    //----------------------------------------------------------------------------------
    always_comb begin:  Controller 
       //-- Default values
       next_state = idle;
       perm_start = 1'b0;
       a_rounds = 1'b1; //-- pa = 12
       bdi_ready = 1'b0;
       ctr_words_rst = 1'b0;
       ctr_words_inc = 1'b0;
       ctr_hash_rst = 1'b0;
       ctr_hash_inc = 1'b0;
       decrypt_rst = 1'b0;
       decrypt_set = 1'b0;
       hash_rst = 1'b0;
       hash_set = 1'b0;
       last_AD_rst = 1'b0;
       last_AD_set = 1'b0;
       no_AD_rst = 1'b0;
       no_AD_set = 1'b0;
       partial_AD_rst = 1'b0;
       partial_AD_set = 1'b0;
       last_M_rst = 1'b0;
       last_M_set = 1'b0;
       no_M_rst = 1'b0;
       no_M_set = 1'b0;
       partial_M_rst = 1'b0;
       partial_M_set = 1'b0;
       tag_check_rst = 1'b0;
       tag_check_set = 1'b0;
       bdo_valid = 1'b0;
       msg_auth = 1'b0;
       msg_auth_valid = 1'b0;
       load_en_sc1 = 1'b0; //--dont care
       
       case (state)
         idle  : begin 
            ctr_words_rst = 1'b1;
            ctr_hash_rst = 1'b1;
            decrypt_rst = 1'b1;
            hash_rst = 1'b1;
            last_AD_rst = 1'b1;
            no_AD_rst = 1'b1;
            partial_AD_rst = 1'b1;
            last_M_rst = 1'b1;
            no_M_rst = 1'b1;
            partial_M_rst = 1'b1;
            tag_check_set = 1'b1; //-- It should be always 1, unless the verification fails
            if (bdi_valid  ==  1'b1) begin
               if (hash_in  ==  1'b1) begin
                  hash_set = 1'b1;
                  perm_start = 1'b1;
                  next_state = Initialization; //-- Initialize the state for hashing
               end else begin
                  next_state = load_Npub; //-- In decryption, skip getting the key and get the nonce
               end
            end else begin
               next_state = idle;
            end
            
         end
         load_key  : begin 
            ctr_words_inc = 1'b1;
            if (ctr_words  ==  2'd3) begin
              ctr_words_rst = 1'b1;
               next_state = wait_Npub;
            end else begin
               next_state = load_key;
            end
            
         end
         wait_Npub  : begin 
            if (bdi_valid  ==  1'b1 && bdi_type  ==  HDR_NPUB) begin
               next_state = load_Npub;
            end else begin
               next_state = wait_Npub;
            end     
         end
         load_Npub  : begin 
            bdi_ready = 1'b1;
            ctr_words_inc = 1'b1;
            if (decrypt_in  ==  1'b1) begin //-- Decryption
               decrypt_set = 1'b1;
            end else begin //-- Encryption
               decrypt_rst = 1'b1;
            end
            if (bdi_eoi  ==  1'b1) begin //-- No AD and no data
               no_AD_set = 1'b1;
               no_M_set = 1'b1;
            end else begin
               no_AD_rst = 1'b1;
               no_M_rst = 1'b1;
            end
            if (bdi_valid  ==  1'b1) begin
               if (ctr_words  ==  2'd3) begin
                  ctr_words_rst = 1'b1;
                  perm_start = 1'b1;
                  next_state = Initialization;
               end else begin
                  next_state = load_Npub;
               end
            end else begin
               ctr_words_inc = 1'b0;
               next_state = load_Npub;
            end
               
         end
         Initialization  : begin 
            if (perm_done  ==  1'b1) begin
               if (hash_reg  ==  1'b1 && bdi_type  ==  HDR_HASH_MSG) begin //-- Absorb message
                  next_state = load_data;
                  load_en_sc1 = 1'b1;
                  
               end else if (no_AD_reg  ==  1'b1 && no_M_reg  ==  1'b1) begin //-- No AD and no data
                  next_state = Finalization;
                  load_en_sc1 = 1'b1;
                  
              end else if (bdi_type  ==  HDR_AD) begin //-- AD
                 next_state = load_AD;
                 load_en_sc1 = 1'b1;
                 
              end else if (bdi_type  ==  HDR_PT || bdi_type  ==  HDR_CT) begin //-- No AD but have data
                 no_AD_set = 1'b1;
                 next_state = load_data;
                 load_en_sc1 = 1'b1;
                 
              end else begin
                 next_state = Initialization_pause;
              end
            end else begin
               next_state = Initialization;
            end
            
         end
         Initialization_pause  : begin 
            if (hash_reg  ==  1'b1 && bdi_type  ==  HDR_HASH_MSG) begin //-- Absorb message
               next_state = load_data;
               load_en_sc1 = 1'b1;
            end else if (no_AD_reg  ==  1'b1 && no_M_reg  ==  1'b1) begin //-- No AD and no data
               next_state = Finalization;
               load_en_sc1 = 1'b1;
            end else if (bdi_type  ==  HDR_AD) begin //-- AD
               next_state = load_AD;
               load_en_sc1 = 1'b1;
            end else if (bdi_type  ==  HDR_PT || bdi_type  ==  HDR_CT) begin //-- No AD but have data
               no_AD_set = 1'b1;
               next_state = load_data;
               load_en_sc1 = 1'b1;
            end else begin
               next_state = Initialization_pause;
               load_en_sc1 = 1'b0;
            end
            
         end
         wait_AD  : begin 
            if (bdi_valid  ==  1'b1) begin
               next_state = load_AD;
            end else begin
               next_state = wait_AD;
            end
            
         end
         load_AD  : begin 
            bdi_ready = 1'b1;
            ctr_words_inc = 1'b1;
            if (bdi_eoi  ==  1'b1) begin //-- No data
               no_M_set = 1'b1;
            end else begin
               no_M_rst = 1'b1;
            end
            if (bdi_eot  ==  1'b1) begin //-- Last block of AD
               last_AD_set = 1'b1;
            end else begin
               last_AD_rst = 1'b1;
            end
            if ((ctr_words  !=  2'd1) || (bdi_size  !=  3'b100)) begin //-- Last partial block
               partial_AD_set = 1'b1;
            end else begin
               partial_AD_rst = 1'b1;
            end
            if (bdi_valid && ctr_words  ==  2'd1 && bdi_size  ==  3'b100) begin //-- Full blocks of AD
               ctr_words_rst = 1'b1;
               perm_start = 1'b1;
               next_state = process_AD;
            end else if (bdi_valid && bdi_eot  ==  1'b1) begin //-- Partial last block of AD
               ctr_words_rst = 1'b1;
               next_state = process_last_AD;
            end else begin
               if (bdi_valid == 1'b0) ctr_words_inc = 1'b0;
               next_state = load_AD;
            end

         end
         process_AD  : begin //-- Except the last block of AD
            a_rounds = 1'b0; //-- pb = 6
            if (perm_done  ==  1'b1) begin
               if (last_AD_reg  ==  1'b1) begin //-- Last full block of AD follows by a 1||0* block
                  next_state = process_last_AD;
               end else begin //-- Still loading AD
                  next_state = wait_AD;
               end
            end else begin
               next_state = process_AD;
            end
            
         end
         process_last_AD  : begin 
            perm_start = 1'b1;
            a_rounds = 1'b0; //-- pb = 6
            if (perm_done  ==  1'b1) begin
               perm_start = 1'b0;
               if (no_M_reg  ==  1'b1) begin //-- No data, go to process tag
                  next_state = Finalization;
               end else begin //-- Done with AD, start loading data
                  next_state = wait_data;
               end
            end else begin
               next_state = process_last_AD;
            end
            
         end
         wait_data  : begin 
            if (bdi_valid  ==  1'b1 && (bdi_type  ==  HDR_PT || bdi_type  ==  HDR_CT)) begin
               next_state = load_data;
            end else begin
               next_state = wait_data;
            end
            
         end
         load_data  : begin 
            bdi_ready = 1'b1;
            ctr_words_inc = 1'b1;
            if (bdi_eot  ==  1'b1) begin //-- Last block of data
               last_M_set = 1'b1;
            end else begin
               last_M_rst = 1'b1;
            end
            if (hash_reg  ==  1'b1 && bdi_size  ==  3'b000) begin //-- Empty message in hash
               no_M_set = 1'b1;
            end else begin
               no_M_rst = 1'b1;
            end
            if (((ctr_words  !=  2'd1) || (bdi_size  !=  3'b100))) begin //-- Partial last block
               partial_M_set = 1'b1;
            end else begin
               partial_M_rst = 1'b1;
            end
            if (hash_reg  ==  1'b0 && bdo_ready  ==  1'b1) begin //-- encryption/Decryption
               bdo_valid = 1'b1;
            end else begin
               bdo_valid = 1'b0;
            end
            if (bdi_valid && (bdi_eot  ==  1'b1 || ctr_words  ==  2'd1)) begin //-- A block of data is received
               ctr_words_rst = 1'b1;
               if (((ctr_words  !=  2'd1) || (bdi_size  !=  3'b100)) && hash_reg  ==  1'b0) begin //-- Encryption/Decryption
                  next_state = Finalization; //-- Partial last block of data does not need a permutation
               end else begin //-- Full blocks of data
                  perm_start = 1'b1;
                next_state = process_data;
               end
            end else begin
              next_state = load_data;
              if (bdi_valid  ==  1'b0) begin
                 ctr_words_inc = 1'b0;                 
              end else begin
                 ctr_words_inc = 1'b1;
              end
            end
            
         end

         process_data  : begin 
            if (hash_reg  ==  1'b0) begin //-- Encryption/Decryption
               a_rounds = 1'b0; //-- pb = 6
            end else begin //-- Hash
               a_rounds = 1'b1; //-- pa = 12
            end
            if (perm_done  ==  1'b1) begin
               if (last_M_reg  ==  1'b1 && hash_reg  ==  1'b1) begin //-- Last block of message in hash
                  if (no_M_reg  ==  1'b0 && partial_M_reg  ==  1'b0) begin //-- Last full block
                     partial_M_set = 1'b1;
                     perm_start = 1'b1;
                     next_state = process_data;
                  end else begin //-- Last partial or 10* block
                     next_state = output_tag;
                  end
               end else if (last_M_reg  ==  1'b1) begin //-- Last full block of data in encryption/decryption
                  next_state = Finalization;
               end else begin //-- Still loading data
                  if (bdi_valid  ==  1'b0) begin
                     next_state = load_data;
                  end else begin
                     next_state = load_data;
                  end
               end
               
            end else begin
               next_state = process_data;
            end
            
         end
         Finalization  : begin 
            perm_start = 1'b1;
            if (perm_done  ==  1'b1) begin
               perm_start = 1'b0;
               if (decrypt_reg  ==  1'b0) begin //-- Encryption
                  next_state = output_tag; //-- TODO: this should goto output_tag_pause if bdo_ready is not high to save a cycle
               end else begin //-- Decryption
                  next_state = wait_tag;
               end
            end else begin
               next_state = Finalization;
            end
            
         end
         output_tag  : begin 
            if (bdo_ready  ==  1'b1) begin
               bdo_valid = 1'b1;
               ctr_words_inc = 1'b1;
               ctr_hash_inc = 1'b1;
               if (hash_reg  ==  1'b0 && ctr_words  ==  2'd3) begin //-- Encryption
                  ctr_words_rst = 1'b1;
                  next_state = idle;
               end else if (hash_reg  ==  1'b1 && ctr_words  ==  2'd1) begin
                  ctr_words_rst = 1'b1;
                  if (ctr_hash  ==  3'd7) begin
                     ctr_hash_rst = 1'b1;
                     next_state = idle;
                  end else begin
                     perm_start = 1'b1;
                     next_state = process_hash;
                  end
               end else begin
                  next_state = output_tag;
               end
            end else begin
               next_state = output_tag;
            end            
         end

         process_hash  : begin 
            if (perm_done  ==  1'b1) begin
               next_state = output_tag;
            end else begin
               next_state = process_hash;
            end
            
         end
       endcase
    end //process 
   
   //-- Datapath
   //--------------------------------------------------------------------------------
   always_comb begin:  Sr_fsm 
      
      Sr = perm_Sr;  //-- Default value
      case (state) //-- Case statement
        idle  : begin 
           if (hash_in  ==  1'b0) begin //-- Encryption/Decryption
              Sr = IV;
           end else begin //-- Hash
              Sr = IV_hash;
           end
        end
        load_AD , load_data  : begin 
           if (ctr_words  ==  2'd0 && bdi_eot  ==  1'b1) begin //-- Partial last block
              if (decrypt_reg  ==  1'b1 && bdi_type  !=  HDR_AD) begin //-- Decrypt
                 if (bdi_valid) Sr = ({(perm_Sr[63:32] ^ pad(bdo_t, bdi_size)),(perm_Sr[31:0])});
              end else begin //-- Encrypt/Hash
                 if (bdi_valid) Sr = ({(perm_Sr[63:32] ^ pad(bdi, bdi_size)),(perm_Sr[31:0])});
              end
              if (bdi_valid && bdi_size  ==  3'd4) begin
                 Sr[31] = perm_Sr[31] ^ 1'b1; //-------
              end else begin
                 Sr[31] = perm_Sr[31];
              end
           end else if (ctr_words  ==  2'd0 && bdi_eot  ==  1'b0) begin //-- Upper 32 bits
              if (decrypt_reg  ==  1'b1 && bdi_type  !=  HDR_AD) begin //-- Decrypt
                 if (bdi_valid) Sr=({(bdi),(perm_Sr[31:0])});
              end else begin //-- Encrypt/Hash
                 if (bdi_valid) Sr = ({((perm_Sr[63:32] ^ bdi)),(perm_Sr[31:0])});
              end
           end else begin //-- Lower 32 bits
              if (decrypt_reg  ==  1'b1 && bdi_type  !=  HDR_AD) begin //-- Decrypt
                  if (bdi_valid) Sr = ({(perm_Sr[63:32]),(perm_Sr[31:0] ^ pad(bdo_t, bdi_size))});
              end else begin //-- Encrypt/Hash
                 if (bdi_valid) Sr = ({(perm_Sr[63:32]),(perm_Sr[31:0] ^ pad(bdi, bdi_size))});
              end
           end
        end
        process_last_AD  : begin 
           if (perm_done  ==  1'b0 && partial_AD_reg  ==  1'b0) begin //-- Last full block of AD follows by a 1||0* block
              Sr = ({(perm_Sr[63] ^ 1'b1),(perm_Sr[62:0])}); //-- p(Sr) xor 1||0*
           end else begin
              Sr = perm_Sr; //-- p(Sr)
           end
        end
        process_data  : begin 
           if (perm_done  ==  1'b1 && last_M_reg  ==  1'b1 && partial_M_reg  ==  1'b0 && hash_reg  ==  1'b1 && no_M_reg  ==  1'b0) begin
              Sr = perm_Sr ^ 64'h8000000000000000; //-- In hash, last full block of message follows by a 10* block and then permutation
           end
        end
            Finalization  : begin 
               if (no_M_reg  ==  1'b1 || (last_M_reg  ==  1'b1 && partial_M_reg  ==  1'b0)) begin //-- No data or full last block
                  Sr = ({(perm_Sr[63] ^ 1'b1),(perm_Sr[62:0])}); //-- p(Sr) xor 1||0*
               end else begin
                  Sr = perm_Sr; //-- p(Sr)
               end
            end
      endcase
   end //process Sr_fsm
   
   //--------------------------------------------------------------------------------
   always_comb begin:  Sc0_fsm 
      
      Sc0 = perm_Sc0; //-- Default value
      case (state) //-- Case statement
        idle  : begin 
           Sc0 = '0;
        end
        load_Npub  : begin 
           Sc0 = key_tag_reg; //-- Key
        end
        Finalization  : begin 
           Sc0 = perm_Sc0 ^ key_tag_reg; //-- p(Sc0) xor key
        end
      endcase
   end //process Sc0_fsm
   
   //--------------------------------------------------------------------------------
   always_comb begin:  Sc1_fsm 
      
      Sc1 = perm_Sc1; //-- Default value
      case (state) //-- Case statement
        idle  : begin 
           Sc1 = '0;
        end
        load_Npub  : begin 
           if (bdi_valid) begin
             Sc1 = ({(perm_Sc1[95:0]),(bdi)}); //-- Nonce
           end         
        end
        Initialization  : begin 
           if ( load_en_sc1  ==  1'b1 && perm_done  ==  1'b1 && hash_reg  ==  1'b0) begin //-- In encryption/decryption, end of initialization
              if (no_AD_reg  ==  1'b1 || bdi_type  !=  HDR_AD) begin //-- NO AD
                 Sc1 = 
                       ({(perm_Sc1[127:1] ^ key_tag_reg[127:1]),(perm_Sc1[0] ^ key_tag_reg[0] ^ 1'b1)}); //-- p(Sc1) xor key xor 0*||1
              end else begin
                 Sc1 = perm_Sc1 ^ key_tag_reg; //-- p(Sc1) xor key
              end
           end else begin
              Sc1 = perm_Sc1;
           end
           
        end
        Initialization_pause  : begin 
           if (load_en_sc1  ==  1'b1 && hash_reg  ==  1'b0) begin //-- In encryption/decryption, end of initialization
              if (no_AD_reg  ==  1'b1 || bdi_type  !=  HDR_AD) begin //-- NO AD
                 Sc1 = 
                       ({(perm_Sc1[127:1] ^ key_tag_reg[127:1]),(perm_Sc1[0] ^ key_tag_reg[0] ^ 1'b1)}); //-- p(Sc1) xor key xor 0*||1
              end else begin
                 Sc1 = perm_Sc1 ^ key_tag_reg; //-- p(Sc1) xor key
              end
           end else begin
              Sc1 = perm_Sc1;
           end
           
        end
        process_last_AD  : begin 
           if (perm_done  ==  1'b1) begin //-- End of AD
              Sc1 = ({(perm_Sc1[127:1]),(perm_Sc1[0] ^ 1'b1)}); //-- p(Sc1) xor 0*||1
              end else begin
                 Sc1 = perm_Sc1;
              end
        end
      endcase
   end //process Sc1_fsm
   //--------------------------------------------------------------------------------
   always_comb begin:  bdo_temp_fsm
      bdo_t = 0; //-- Default value
      case (state) //-- Case statement
        load_data  : begin 
           if (hash_reg  ==  1'b0) begin //-- Encryption/Decryption
              if (ctr_words  ==  2'd0) begin //-- Upper 32 bits
                 bdo_t = trunc(perm_Sr[63:32] ^ bdi, bdi_eot, bdi_size);
              end else begin //-- Lower 32 bits
                 bdo_t = trunc(perm_Sr[31:0] ^ bdi, bdi_eot, bdi_size);
              end
           end
        end
        output_tag  : begin 
           if (hash_reg  ==  1'b0) begin //-- Encryption/Decryption
              bdo_t = ext(Sc1 ^ key_tag_reg, ctr_words);
           end else begin //-- Hash
              if (ctr_words  ==  2'd0) begin //-- Upper 32 bits
                 bdo_t = perm_Sr[63:32];
              end else begin //-- Lower 32 bits
                 bdo_t = perm_Sr[31:0];
              end
           end
        end
      endcase
   end //process bdo_temp_fsm
   //--------------------------------------------------------------------------------
   always_comb begin:  end_of_block_fsm       
      end_of_block = 1'b0; //-- Default value
      case (state) //-- Case statement
        load_data  : begin 
           if (hash_reg  ==  1'b0) begin //-- Encryption/Decryption
              end_of_block = bdi_eot;
           end else begin
              end_of_block = 1'b0;
           end
        end
        output_tag  : begin 
           if (ctr_words  ==  2'd3 || ctr_hash  ==  3'd7) begin //-- Last word of tag or hash
              end_of_block = 1'b1;
           end else begin
              end_of_block = 1'b0;
           end
        end
      endcase
   end //process 
   
endmodule //architecture Behavioral
