`timescale 1ns / 1ps
//-- ASCON_AEAD wrapper for  hw root of trust
// Efforts from Ryan Robucci, Vajesh Mistry, Saurin Shah, Chintan Patel
// academic research effort, refactoring and verification underway, code for reference only, untested
  
module ascon_wrapper(clk,
                     rst,
                     bdi,
                     bdi_valid,
                     bdi_ready,
                     bdi_pad_loc,
                     bdi_valid_bytes,
                     bdi_size,
                     bdi_eot,
                     bdi_eoi,
                     bdi_type,
                     hash_in,
                     decrypt_in,
                     bdo,
                     bdo_valid,         //use as handshake
                     bdo_we,            //use as WE
                     bdo_ready,
                     bdo_valid_bytes,
                     end_of_block,
                     bdo_type,
                     msg_auth,
                     msg_auth_valid,
                     msg_auth_ready,
                     pcr_update,
                     ctrl_opcode,
                     key_tag_reg,
                     bus_reg_pcr_de,
                     bus_reg_pcr_d,
                     bus_reg_pcr_q 
);


   localparam ROT_HDR_AD   = 4'b0001;
   localparam ROT_HDR_NPUB = 4'b1101;
   localparam ROT_HASH_MSG = 4'b0111;

   typedef enum logic [3:0] {
                             OP_idle    = 4'd0,
                             OP_load_PS = 4'd1,
                             OP_load_AK = 4'd2,
                             OP_hashing = 4'd3,
                             OP_extend  = 4'd4,
                             OP_encrypt = 4'd5,
                             OP_decrypt = 4'd6,
                             OP_derive  = 4'd7,
                             OP__derive = 4'd8,
                             OP_quote   = 4'd9,
                             OP_xtend   = 4'd11
                             } opcode_t;

   typedef enum int {
                     S_idle      ,
                     S_load_AK_00,
                     S_load_AK_01,
                     S_load_AK_02,
                     S_load_AK_03,
                     S_load_AK_04,
                     S_load_AK_05,
                     S_load_AK_06,
                     S_load_AK_07,
                     S_load_AK_08,
                     S_load_AK_09,
                     S_load_AK_40,
                     S_load_AK_41,
                     S_load_AK_50,
                     S_load_AK_51,
                     S_load_AK_52,
                     S_load_AK_53,
                     S_load_AK_54,
                     S_load_AK_55,
                     S_load_AK_56,
                     S_load_AK_57,
                     S_Extend_00 ,
                     S_Extend_01 ,
                     S_Extend_02 ,
                     S_Extend_03 ,
                     S_Extend_04 ,
                     S_Extend_05 ,
                     S_Extend_06 ,
                     S_Extend_07 ,
                     S_Extend_40 ,
                     S_Extend_41 ,
                     S_Extend_020,
                     S_Extend_040,
                     S_Extend_060,
                     S_Extend_50 ,
                     S_derive_000,
                     S_derive_100,
                     S_derive_101,
                     S_derive_200,
                     S_derive_201,
                     S_derive_202,
                     S_derive_203,
                     S_derive_204,
                     S_derive_205,
                     S_derive_206,
                     S_derive_207 
} state_t;  //use int to allow flexibility for state encoding during synthesis


   input                clk,rst;
   input [31:0]         bdi;
   input                bdi_valid;
   input [3:0]          bdi_pad_loc,bdi_valid_bytes;
   input [2:0]          bdi_size;
   input                bdi_eoi,bdi_eot;
   input [3:0]          bdi_type;
   input                hash_in, decrypt_in;
   input                bdo_ready;
   input                msg_auth_ready;
   output logic         bdi_ready;
   output [31:0]        bdo;
   output logic         bdo_valid;   
   output logic         bdo_we;
   output               end_of_block; //todo use this in sw
   output [3:0]         bdo_valid_bytes, bdo_type;
   output logic         msg_auth,msg_auth_valid;
   input opcode_t     ctrl_opcode;
   input                pcr_update;
   input [127:0]        key_tag_reg;
   output logic         bus_reg_pcr_de;
   output logic [255:0] bus_reg_pcr_d;
   input [255:0]        bus_reg_pcr_q;  //This has the true value




   logic                hash_in_ascon;
   logic                bdo_valid_ascon;
      


   
   logic                bdi_ready_ascon;
   logic [3:0]          bdi_valid_bytes_ascon;   
   logic                bdo_ready_ascon;                
   logic                bdi_valid_ascon;
   logic                bdi_eot_ascon;
   logic                bdi_eoi_ascon;                
   logic [3:0]          bdi_type_ascon;
   logic [2:0]          bdi_size_ascon;
   
   
   state_t CS,NS;
  
   wire [255:0]         pcr = bus_reg_pcr_q;
   logic [31:0]         bdi_ascon;
   logic [127:0]        ps_reg;   
   assign ps_reg="A 16 byte secret";
   logic [127:0]        ak_reg;  //TODO: initial value 0 was provided here in previous implementation test, investigate effect of removal
   logic [127:0]        key_tag_mux;

   reg                  bdi_valid_prev;
   reg                  bdo_ready_prev;
   logic                ak_en;
   wire                 msg_auth_ascon;

   logic                bdi_token_ready;   
   logic                bdi_token_fresh;
   logic                bdi_token_available;
   logic                bdi_token_full_next;
   logic                bdi_tokens_stored,_bdi_tokens_stored;   

   //glue logic
   logic                null_op;
   assign null_op = (ctrl_opcode == OP_idle);
   logic null_op_prev;
   logic new_op;
   assign new_op = !null_op & null_op_prev;   
   always_ff @ (posedge clk) begin
      null_op_prev<=null_op;      
   end

   
  //  BEGIN BDI GLUE LOGIC  

  always @(posedge clk) begin
      bdi_valid_prev  <= bdi_valid;      
   end

   assign bdi_ready = !bdi_token_full_next;   
        
   always_comb begin
      bdi_token_fresh = bdi_valid && (!bdi_valid_prev) && (!rst); //TODO:consider need for rst
   end 
   
   always @ (posedge clk) begin
      if (rst) begin
         bdi_tokens_stored <= 1'b0;
      end else begin         
         bdi_tokens_stored<=_bdi_tokens_stored;
      end
   end 
   
   always_comb begin
      case (bdi_tokens_stored)
        1'b0: begin
           bdi_token_available = bdi_token_fresh;
           _bdi_tokens_stored   = bdi_token_fresh && ! bdi_token_ready;
           bdi_token_full_next = _bdi_tokens_stored;           
        end
        
        1'b1: begin
           bdi_token_available = 1'b1;
           _bdi_tokens_stored   = bdi_token_fresh || ! bdi_token_ready;
           bdi_token_full_next = _bdi_tokens_stored;           
        end
      endcase
   end // always_comb

   // END BDI GLUE LOGIC  

   // BEGIN BDO GLUE LOGIC
   logic  bdo_ready_handshake, bdo_ready_handshake_prev;   
   logic  bdo_valid_sm_prev,bdo_valid_sm,bdo_valid_sm_fresh;
   logic  bdo_valid_prev;
   assign bdo_we = bdo_valid_sm;   
   
   logic bdo_ready_fresh;
   assign bdo_ready_fresh    = bdo_ready    ==1'b1    && bdo_ready_prev     == 1'b0;
   assign bdo_valid_sm_fresh = bdo_valid_sm ==1'b1    && bdo_valid_sm_prev  == 1'b0;
   
   always_ff @(posedge clk)begin
      //TODO: consider sysreset for glue logic
      bdo_ready_handshake_prev <=bdo_ready_handshake;
      bdo_valid_sm_prev        <=bdo_valid_sm;
      bdo_valid_prev           <=bdo_valid;
      bdo_ready_prev           <= bdo_ready;
   end

   always_comb begin
      if (bdo_ready_fresh && !bdo_valid_sm_fresh) begin
         bdo_valid = 1'b0;
      end else if (bdo_valid_sm_fresh) begin
         bdo_valid = 1'b1;
      end else begin
         bdo_valid = bdo_valid_prev;         
      end      
   end

   always_comb begin
      if (bdo_ready_fresh) begin
         bdo_ready_handshake=1; //fresh asserts alway cause a high
      end else  if (bdo_valid_sm_prev==1'b1) begin
         bdo_ready_handshake=0; //assume bdo was consumed, but really it was written to the bus register, ascon should not update it until the processor says that it is ready again
      end else begin
         bdo_ready_handshake = bdo_ready_handshake_prev;
      end         
   end // always_comb
   
   // END BDO GLUE LOGIC

   
   always_comb begin
      bus_reg_pcr_d = {pcr[(255-32):0],bdo};   
   end   
     
   always_ff @(posedge clk) begin
      if(ak_en) 
        begin
           ak_reg <= {ak_reg[(127-32):0],bdo};
      end
      else ak_reg <= ak_reg;
   end
   
   always_ff @(posedge clk) begin
      if(rst || (bdi_type == 4'd13)) begin 
         msg_auth <= 0;
      end else begin
         if(msg_auth_ascon) msg_auth <= 1;
         else msg_auth <= msg_auth;
      end
   end      


   always_ff @(posedge clk) begin
      if (rst==1'b1) begin
         CS<=S_idle;
      end else begin
         CS<=NS;
      end
   end

   typedef enum logic [1:0] {
                             SEL_KEY_TAG,
                             SEL_PS,
                             SEL_AK,
                             SEL_ZERO
                        } key_tag_mux_sel_t;
   
   key_tag_mux_sel_t key_tag_sel;
   key_tag_mux_sel_t key_tag_sel_prev;
   
   always_ff @(posedge clk) begin
      if (rst) begin
         key_tag_sel_prev <= SEL_KEY_TAG;
      end else begin
         key_tag_sel_prev <= key_tag_sel;
      end
   end

   always_comb begin
      case (key_tag_sel)
        SEL_KEY_TAG: key_tag_mux = key_tag_reg;
        SEL_PS     : key_tag_mux = ps_reg;
        SEL_AK     : key_tag_mux = ak_reg;
        SEL_ZERO   : key_tag_mux = '0; // not functionally used
      endcase
   end

   always_comb begin
      ak_en                 =  '0;
      NS                    =  CS;
      bdi_ascon             =  bdi;      
      hash_in_ascon         =  hash_in;      
      bdi_valid_bytes_ascon =  bdi_valid_bytes ;
      bdi_valid_ascon       =  bdi_token_available;                
      bdi_eot_ascon         =  bdi_eot;                
      bdi_eoi_ascon         =  bdi_eoi;                
      bdi_type_ascon        =  bdi_type;
      bdo_ready_ascon       =  bdo_ready_handshake;
      bdi_size_ascon        =  bdi_size;
      bdi_token_ready       =  bdi_ready_ascon;      
      bdo_valid_sm          =  bdo_valid_ascon;
      bus_reg_pcr_de        =  '0; //only written in final extend states
      key_tag_sel           =  key_tag_sel_prev; // key_tag_mux <= key_tag_mux;

      case (CS)
        S_idle: begin
           case(ctrl_opcode)
             OP_idle,OP_load_PS:   begin
                key_tag_sel = key_tag_sel_prev; // key_tag_mux <= key_tag_mux;
             end
             OP_load_AK :   begin
                key_tag_sel = SEL_PS; 
                if (new_op) begin
                  NS=S_load_AK_00;
                end
             end
             OP_hashing :   begin
                key_tag_sel = SEL_KEY_TAG;
             end            
             OP_extend  :   begin
                key_tag_sel = SEL_KEY_TAG;
             end
             OP_encrypt :   begin
                key_tag_sel = SEL_KEY_TAG;
             end           
             OP_decrypt :   begin
                key_tag_sel = SEL_KEY_TAG;
             end
             OP_derive  :   begin
                key_tag_sel = SEL_PS; 
             end
             OP__derive :   begin
                key_tag_sel = SEL_PS;
                NS=S_derive_000;
             end
             OP_quote   :   begin
                key_tag_sel = SEL_AK;
             end
             OP_xtend  :   begin
                key_tag_sel = SEL_KEY_TAG;
                   NS=S_Extend_00;
             end
             default :   begin //TODO address default
                key_tag_sel = SEL_ZERO; //key_tag presented as 0
             end
          endcase
        end // S_idle



        //load empty nonce 4x4 0
        S_load_AK_00: begin
           bdi_type_ascon=ROT_HDR_NPUB;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_ascon=0; //this is an empty nonce           
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_01;
           end           
        end

        
        S_load_AK_01: begin
           bdi_type_ascon=ROT_HDR_NPUB;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS; 
           bdi_valid_ascon=1;
           bdi_ascon=0; //this is an empty nonce           
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_02;
           end           
        end
        S_load_AK_02: begin
           bdi_type_ascon=ROT_HDR_NPUB;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS; //key_tag_mux <= PS_reg;
           bdi_valid_ascon=1;
           bdi_ascon=0; //this is an empty nonce           
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_03;
           end           
        end
        S_load_AK_03: begin
           bdi_type_ascon=ROT_HDR_NPUB;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_ascon=0; //this is an empty nonce
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;           
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_04;
           end           
        end

        S_load_AK_04: begin
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_ascon="-Get";
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_05;
           end           
        end
        
        S_load_AK_05: begin
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_ascon="-AK-";            
           bdi_eot_ascon =1;
           bdi_eoi_ascon=1;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_06;
           end           
        end
        
        
        S_load_AK_06: begin
           bdo_ready_ascon=1;           
           if (bdo_valid_ascon) begin
              NS=S_load_AK_07;
              ak_en=1;
           end           
        end

        S_load_AK_07: begin
           bdo_ready_ascon=1;           
           if (bdo_valid_ascon) begin
              NS=S_load_AK_08;
              ak_en=1;
           end           
        end
        
        S_load_AK_08: begin
           bdo_ready_ascon=1;           
           if (bdo_valid_ascon) begin
              NS=S_load_AK_09;
              ak_en=1;
           end           
        end

        S_load_AK_09: begin
           bdo_ready_ascon=1;           
           if (bdo_valid_ascon) begin
              NS=S_load_AK_40;
              ak_en=1;
           end           
        end

        //BY NOW THE AK IS SAVED, TIME to go again with opcode quote which uses key sel ak

        S_load_AK_40: begin
           if (bdi_type == ROT_HDR_NPUB) begin
              //bdi_token_ready=bdi_ready_ascon;              
              if (bdi_ready_ascon) begin
                NS=S_load_AK_41;
              end
           end 
           bdi_type_ascon=ROT_HDR_NPUB; //FORCE ROT_HDR_NPUB
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0; // do not allow SW to end interaction            
        end
        
        //states to enforce NPUB   TODO:maybe count
        S_load_AK_41: begin
           if (bdi_type != ROT_HDR_NPUB) begin //todo, use eot instead?
              NS=S_load_AK_50; 
           end
           bdi_type_ascon=ROT_HDR_NPUB; //FORCE ROT_HDR_NPUB
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_eoi_ascon=0; // do not allow SW to end interaction  NOTE:1
        end


        S_load_AK_50: begin
           bdi_ascon = pcr[255:224];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK; //key_tag_mux <= PS_reg;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdi_token_ready=0; //TODO: check anywhere for which the token control passthrough is one-way, this seems to be the cause of problems.  This was based on the previous understanding of the protocol.           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_51;
           end           
        end

        S_load_AK_51: begin
           bdi_ascon = pcr[223:192];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK; //key_tag_mux <= PS_reg;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_52;
           end           
        end

        S_load_AK_52: begin
           bdi_ascon = pcr[191:160];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_53;
           end           
        end

        S_load_AK_53: begin
           bdi_ascon = pcr[159:128];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_54;
           end           
        end

        S_load_AK_54: begin
           bdi_ascon = pcr[127:96];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_55;
           end           
        end

        S_load_AK_55: begin
           bdi_ascon = pcr[95:64];           
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_56;
           end           
        end

        S_load_AK_56: begin
           bdi_ascon = pcr[63:32];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_load_AK_57;
           end           
        end

        S_load_AK_57: begin
           bdi_ascon = pcr[31:0];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =1;
           bdi_eoi_ascon=1;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_idle; //result bdo tobe read by software
           end           
        end

        S_Extend_00: begin
           bdi_ascon = pcr[255:224];
           bdi_type_ascon=ROT_HASH_MSG;
           hash_in_ascon = 1;      
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_01;
           end           
        end


        S_Extend_01: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[223:192];
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;  
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_020;
           end                
        end

        S_Extend_020: begin
           hash_in_ascon = 1;      
           bdi_token_ready=0;           
           if (!bdi_ready_ascon) begin
              NS=S_Extend_02;
           end
        end

        S_Extend_02: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[191:160];
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_03;
           end           
        end

        S_Extend_03: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[159:128];
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_040;
           end           
        end

        S_Extend_040: begin
           hash_in_ascon = 1;      
           bdi_token_ready=0;           
           if (!bdi_ready_ascon) begin
              NS=S_Extend_04;
           end
        end

        S_Extend_04: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[127:96];
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_05;
           end           
        end

        S_Extend_05: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[95:64];           
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_060;
           end           
        end


        S_Extend_060: begin
           hash_in_ascon = 1;
           bdi_token_ready=0;
           if (!bdi_ready_ascon) begin
              NS=S_Extend_06;
           end
        end
        
        S_Extend_06: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[63:32];
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_07;
           end           
        end


        S_Extend_07: begin
           hash_in_ascon = 1;      
           bdi_ascon = pcr[31:0];
           bdi_type_ascon=ROT_HASH_MSG;
           bdi_size_ascon=4;
           key_tag_sel = SEL_AK;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_Extend_40; //result bdo to be read by software
           end
        end


        //TODO: add exception or make software reset to recover
        //states to enforce HASH
        S_Extend_40: begin
           hash_in_ascon = 1;
           bdi_type_ascon=ROT_HASH_MSG; //FORCE ROT_HDR_NPUB           
           if (bdi_type == ROT_HASH_MSG && bdi_valid && bdi_ready_ascon) begin //todo: hash_in?
              NS=S_Extend_41;
              bdi_token_ready = 1;
           end else begin
              bdi_token_ready=0;       
              bdi_valid_ascon=0;
              bdi_eot_ascon =0;
              bdi_eoi_ascon=0;
           end
        end


        
        S_Extend_41: begin
           hash_in_ascon = 1;
           bdi_type_ascon=ROT_HASH_MSG; //FORCE ROT_HDR_NPUB           
           if (bdi_type == ROT_HASH_MSG && bdi_valid && bdi_ready_ascon ) begin //todo: hash_in? what about bdi eoi?
              if (bdi_eot) begin // bdi_eoi
                 NS=S_Extend_50;
              end
              bdi_token_ready = 1;
           end //todo exeption handling? such as other types?
        end
        
        S_Extend_50: begin //TODO: can eom be used here?
           bdo_ready_ascon=1;           
           if (bdo_valid_ascon) begin 
              bus_reg_pcr_de = 1;
           end
           if (end_of_block) begin
              NS = S_idle;
           end
        end


        
        //TODO: states to enforce bdo?



        // Begin DERIVE


        //TODO: add exception or make software reset to recover
        //states to enforce NPUB        
        S_derive_000: begin
           if (bdi_type == ROT_HDR_NPUB && bdi_ready_ascon==1'b1) begin //NPUB
              //bdi_ready=1; //TODO:consider this once more, is handshake ok              
              if (bdi_eot==1'b1) begin //NPUB
                  NS=S_derive_100;
              end              
           end
           bdi_type_ascon=ROT_HDR_NPUB; //FORCE ROT_HDR_NPUB
           //bdi_size_ascon=4; //TODO: not sure about the length of the nonce assuming 4 and 4
           key_tag_sel = SEL_PS;
        end
        
        S_derive_100: begin
           bdi_type_ascon=ROT_HDR_AD; //ROT_HDR_AD
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_ascon="Exte";
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           bdo_valid_sm = 0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_101;
           end           
        end
        
        S_derive_101: begin
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS; 
           bdi_valid_ascon=1;
           bdi_ascon="rnal";            
           bdi_eot_ascon = 0;
           bdi_eoi_ascon = 0;
           bdo_ready_ascon=0;
           bdi_token_ready=0;           
           bdo_valid_sm = 0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_200;
           end           
        end

        S_derive_200: begin
           bdi_ascon = pcr[255:224];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdi_token_ready=0;           
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_201;
           end           
        end

        S_derive_201: begin
           bdi_ascon = pcr[223:192];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_202;
           end           
        end

        S_derive_202: begin
           bdi_ascon = pcr[191:160];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_203;
           end           
        end


        S_derive_203: begin
           bdi_ascon = pcr[159:128];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_204;
           end           
        end

        S_derive_204: begin
           bdi_ascon = pcr[127:96];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_205;
           end           
        end

        S_derive_205: begin
           bdi_ascon = pcr[95:64];           
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;           
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_206;
           end           
        end

        S_derive_206: begin
           bdi_ascon = pcr[63:32];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon =0;
           bdi_eoi_ascon=0;            
           bdo_ready_ascon=0;           
           bdi_token_ready=0;
           bdo_valid_sm=0;           
           if (bdi_ready_ascon) begin
              NS=S_derive_207;
           end           
        end

        S_derive_207: begin
           bdi_ascon = pcr[31:0];
           bdi_type_ascon=ROT_HDR_AD;
           bdi_size_ascon=4;
           key_tag_sel = SEL_PS;
           bdi_valid_ascon=1;
           bdi_eot_ascon=1;
           bdi_eoi_ascon=1;            
           bdo_ready_ascon=0;           
           bdo_valid_sm=0;           
           bdi_token_ready=0;           
           if (bdi_ready_ascon) begin
              NS=S_idle; //result bdo to be read by software after sw sends remaining ad
           end    
        end

        
        // END DERIVE
        
     endcase        
   end // always_ff @ (posedge clk,rst) 


   
   ASCON128 a1 (.clk(clk),
                .rst(rst),
                .bdi(bdi_ascon),
                .bdi_valid(bdi_valid_ascon),
                .bdi_ready(bdi_ready_ascon),
                .bdi_pad_loc(bdi_pad_loc),
                .bdi_valid_bytes(bdi_valid_bytes_ascon),
                .bdi_size(bdi_size_ascon),
                .bdi_eot(bdi_eot_ascon),
                .bdi_eoi(bdi_eoi_ascon),
                .bdi_type(bdi_type_ascon),
                .hash_in(hash_in_ascon),
                .decrypt_in(decrypt_in),
                .bdo(bdo),
                .bdo_valid(bdo_valid_ascon),
                .bdo_ready(bdo_ready_ascon),
                .bdo_valid_bytes(bdo_valid_bytes),
                .end_of_block(end_of_block),
                //removed -- .bdo_type(bdo_type),
                .msg_auth(msg_auth_ascon),
                .msg_auth_valid(msg_auth_valid),
                .msg_auth_ready(msg_auth_ready),
                .key_tag_reg(key_tag_mux)
                );
endmodule
