// ascon-based root of trust for opentitan platform
// Efforts from Ryan Robucci, Vajesh Mistry, Saurin Shah, Chintan Patel
// academic research effort, refactoring and verification underway, code for reference only, untested

module rot(
           input  clk_i,
           input  rst_ni,
           
           input  tlul_pkg::tl_h2d_t tl_i,
           output tlul_pkg::tl_d2h_t tl_o,

           // Config
           input  devmode_i // NOTE: UNUSED , intent: if 1, explicit error return for unmapped register access
           );
   
   import    rot_reg_pkg::*;
   import    tlul_pkg::*;
   
   rot_reg2hw_t reg2hw;
   rot_hw2reg_t hw2reg;
   
   rot_reg_top reg_top_instance (
                                 .clk_i,
                                 .rst_ni,
                                 
                                 .tl_i,
                                 .tl_o,
                                 
                                 // To HW
                                 .reg2hw, // Write
                                 .hw2reg, // Read
                                 
                                 // Config
                                 .devmode_i(1'b1)
                                 );
   
   assign hw2reg.stat.key_ready.de      = 1;
   assign hw2reg.stat.bdi_ready.de      = 1;
   assign hw2reg.stat.bdo_valid.de      = 1;
   assign hw2reg.stat.end_of_block.de   = 1;
   assign hw2reg.stat.msg_auth.de       = 1;
   assign hw2reg.stat.msg_auth_valid.de = 1;
   assign reg2hw.stat.key_ready.qe      = 1;
   assign reg2hw.stat.bdi_ready.qe      = 1;
   assign reg2hw.stat.bdo_valid.qe      = 1;
   assign reg2hw.stat.end_of_block.qe   = 1;
   assign reg2hw.stat.msg_auth.qe       = 1;
   assign reg2hw.stat.msg_auth_valid.qe = 1;
   
   logic             bdo_we;
   logic             bus_reg_pcr_de;
   logic [255:0]     bus_reg_pcr_e,bus_reg_pcr_q;
   //logic             ctrlReset;
   logic             asconReset;
   logic [127:0]     key_tag_reg_128b;

   assign key_tag_q_128b = {{reg2hw.key[0].q[7:0],reg2hw.key[0].q[15:8],reg2hw.key[0].q[23:16],reg2hw.key[0].q[31:24]},
					                  {reg2hw.key[1].q[7:0],reg2hw.key[1].q[15:8],reg2hw.key[1].q[23:16],reg2hw.key[1].q[31:24]},
					                  {reg2hw.key[2].q[7:0],reg2hw.key[2].q[15:8],reg2hw.key[2].q[23:16],reg2hw.key[2].q[31:24]},
					                  {reg2hw.key[3].q[7:0],reg2hw.key[3].q[15:8],reg2hw.key[3].q[23:16],reg2hw.key[3].q[31:24]}};
   
   
   //assign ctrlReset = reg2hw.ctrl.rst_sync.q; //TODO: note in documentation as unused 
   assign asconReset = ~rst_ni;
   
   assign hw2reg.bdo.de = bdo_we;
   assign reg2hw.bdo.qe = 1;
   
   assign hw2reg.pcr[7].de =bus_reg_pcr_de;
   assign hw2reg.pcr[6].de =bus_reg_pcr_de;
   assign hw2reg.pcr[5].de =bus_reg_pcr_de;
   assign hw2reg.pcr[4].de =bus_reg_pcr_de;
   assign hw2reg.pcr[3].de =bus_reg_pcr_de;
   assign hw2reg.pcr[2].de =bus_reg_pcr_de;
   assign hw2reg.pcr[1].de =bus_reg_pcr_de;
   assign hw2reg.pcr[0].de =bus_reg_pcr_de;


   // flip bytes: last-minute design change to conform with interface specification for the order of the bytes
   function [255:0] endianswapineachword;
      input [255:0]  x;
      endianswapineachword = {x[  7:  0] ,x[ 15:  8] ,x[ 23: 16] ,x[ 31: 24],
                   x[ 39: 32] ,x[ 47: 40] ,x[ 55: 48] ,x[ 63: 56],
                   x[ 71: 64] ,x[ 79: 72] ,x[ 87: 80] ,x[ 95: 88],
                   x[103: 96] ,x[111:104] ,x[119:112] ,x[127:120],
                   x[135:128] ,x[143:136] ,x[151:144] ,x[159:152],
                   x[167:160] ,x[175:168] ,x[183:176] ,x[191:184],
                   x[199:192] ,x[207:200] ,x[215:208] ,x[223:216],
                   x[231:224] ,x[239:232] ,x[247:240] ,x[255:248]};
   endfunction
   
   assign {hw2reg.pcr[7].d,
           hw2reg.pcr[6].d,
           hw2reg.pcr[5].d,
           hw2reg.pcr[4].d,
           hw2reg.pcr[3].d,
           hw2reg.pcr[2].d,
           hw2reg.pcr[1].d,
           hw2reg.pcr[0].d
           } = endianswapineachword(bus_reg_pcr_d);
   
   assign bus_reg_pcr_q = endianswapineachword({reg2hw.pcr[7].q,
                                      reg2hw.pcr[6].q,
                                      reg2hw.pcr[5].q,
                                      reg2hw.pcr[4].q,
                                      reg2hw.pcr[3].q,
                                      reg2hw.pcr[2].q,
                                      reg2hw.pcr[1].q,
                                      reg2hw.pcr[0].q
                                      });

   ascon_wrap UUT (.clk               (clk_i),                                                           
                   .rst               (asconReset),
                   .bdi               ({reg2hw.bdi.q[7:0],reg2hw.bdi.q[15:8],reg2hw.bdi.q[23:16],reg2hw.bdi.q[31:24]}),
                   .bdi_valid         (reg2hw.ctrl.bdi_valid.q),
                   .bdi_ready         (hw2reg.stat.bdi_ready.d),
                   .bdi_pad_loc       (4'b0),                             //assume always working with valid 4-byte chunks
                   .bdi_valid_bytes   (4'b1111),                          //assume always working with valid 4-byte chunks
                   .bdi_size          (reg2hw.ctrl.bdi_size.q),           //TODO: look at the reset value !! - vraj
                   .bdi_eot           (reg2hw.ctrl.bdi_eot.q),
                   .bdi_eoi           (reg2hw.ctrl.bdi_eoi.q),
                   .bdi_type          (reg2hw.ctrl.bdi_type.q),
                   .hash_in           (reg2hw.ctrl.hash_in.q),
                   .decrypt_in        (reg2hw.ctrl.decrypt_in.q),         // NOTE: set to 0 for derive
                   .bdo               ({hw2reg.bdo.d[7:0],hw2reg.bdo.d[15:8],hw2reg.bdo.d[23:16],hw2reg.bdo.d[31:24]}),
                   .bdo_valid         (hw2reg.stat.bdo_valid.d),
                   .bdo_we            (bdo_we),
                   .bdo_ready         (reg2hw.ctrl.bdo_ready.q),
                   .bdo_valid_bytes   (),                                 // NOTE: unused, simply deterministic
                   .end_of_block      (hw2reg.stat.end_of_block.d),
                   .bdo_type          (),                                 // NOTE: unused, simply deterministic
                   .msg_auth          (hw2reg.stat.msg_auth.d),
                   .msg_auth_valid    (hw2reg.stat.msg_auth_valid.d),
                   .msg_auth_ready    (reg2hw.ctrl.msg_auth_ready.q),     // TODO: need to change this for enc/dec
                   .pcr_update        (reg2hw.ctrl.pcr_update.q),
		               .opcode            (reg2hw.ctrl.opcode.q),
                   .key_tag_reg       (key_tag_q_128b),
		               .bus_reg_pcr_de    (bus_reg_pcr_de),
		               .bus_reg_pcr_d     (bus_reg_pcr_d),
		               .bus_reg_pcr_q     (bus_reg_pcr_q)
                   );

endmodule

