// Copyright lowRISC contributors.
// Licensed under the Apache License, Version 2.0, see LICENSE for details.
// SPDX-License-Identifier: Apache-2.0
//
// Register Top module auto-generated by `reggen`

`include "prim_assert.sv"

module rot_reg_top (
  input clk_i,
  input rst_ni,

  // Below Regster interface can be changed
  input  tlul_pkg::tl_h2d_t tl_i,
  output tlul_pkg::tl_d2h_t tl_o,
  // To HW
  output rot_reg_pkg::rot_reg2hw_t reg2hw, // Write
  input  rot_reg_pkg::rot_hw2reg_t hw2reg, // Read

  // Config
  input devmode_i // If 1, explicit error return for unmapped register access
);

  import rot_reg_pkg::* ;

  localparam int AW = 6;
  localparam int DW = 32;
  localparam int DBW = DW/8;                    // Byte Width

  // register signals
  logic           reg_we;
  logic           reg_re;
  logic [AW-1:0]  reg_addr;
  logic [DW-1:0]  reg_wdata;
  logic [DBW-1:0] reg_be;
  logic [DW-1:0]  reg_rdata;
  logic           reg_error;

  logic          addrmiss, wr_err;

  logic [DW-1:0] reg_rdata_next;

  tlul_pkg::tl_h2d_t tl_reg_h2d;
  tlul_pkg::tl_d2h_t tl_reg_d2h;

  assign tl_reg_h2d = tl_i;
  assign tl_o       = tl_reg_d2h;

  tlul_adapter_reg #(
    .RegAw(AW),
    .RegDw(DW)
  ) u_reg_if (
    .clk_i,
    .rst_ni,

    .tl_i (tl_reg_h2d),
    .tl_o (tl_reg_d2h),

    .we_o    (reg_we),
    .re_o    (reg_re),
    .addr_o  (reg_addr),
    .wdata_o (reg_wdata),
    .be_o    (reg_be),
    .rdata_i (reg_rdata),
    .error_i (reg_error)
  );

  assign reg_rdata = reg_rdata_next ;
  assign reg_error = (devmode_i & addrmiss) | wr_err ;

  // Define SW related signals
  // Format: <reg>_<field>_{wd|we|qs}
  //        or <reg>_{wd|we|qs} if field == 1 or 0
  logic [31:0] pcr0_qs;
  logic [31:0] pcr1_qs;
  logic [31:0] pcr2_qs;
  logic [31:0] pcr3_qs;
  logic [31:0] pcr4_qs;
  logic [31:0] pcr5_qs;
  logic [31:0] pcr6_qs;
  logic [31:0] pcr7_qs;
  logic [31:0] key0_qs;
  logic [31:0] key0_wd;
  logic key0_we;
  logic [31:0] key1_qs;
  logic [31:0] key1_wd;
  logic key1_we;
  logic [31:0] key2_qs;
  logic [31:0] key2_wd;
  logic key2_we;
  logic [31:0] key3_qs;
  logic [31:0] key3_wd;
  logic key3_we;
  logic ctrl_key_valid_qs;
  logic ctrl_key_valid_wd;
  logic ctrl_key_valid_we;
  logic ctrl_bdi_valid_qs;
  logic ctrl_bdi_valid_wd;
  logic ctrl_bdi_valid_we;
  logic [3:0] ctrl_bdi_type_qs;
  logic [3:0] ctrl_bdi_type_wd;
  logic ctrl_bdi_type_we;
  logic [2:0] ctrl_bdi_size_qs;
  logic [2:0] ctrl_bdi_size_wd;
  logic ctrl_bdi_size_we;
  logic ctrl_bdo_ready_qs;
  logic ctrl_bdo_ready_wd;
  logic ctrl_bdo_ready_we;
  logic ctrl_rst_sync_qs;
  logic ctrl_rst_sync_wd;
  logic ctrl_rst_sync_we;
  logic ctrl_hash_in_qs;
  logic ctrl_hash_in_wd;
  logic ctrl_hash_in_we;
  logic ctrl_decrypt_in_qs;
  logic ctrl_decrypt_in_wd;
  logic ctrl_decrypt_in_we;
  logic ctrl_pcr_update_qs;
  logic ctrl_pcr_update_wd;
  logic ctrl_pcr_update_we;
  logic ctrl_bdi_eot_qs;
  logic ctrl_bdi_eot_wd;
  logic ctrl_bdi_eot_we;
  logic ctrl_bdi_eoi_qs;
  logic ctrl_bdi_eoi_wd;
  logic ctrl_bdi_eoi_we;
  logic [3:0] ctrl_opcode_qs;
  logic [3:0] ctrl_opcode_wd;
  logic ctrl_opcode_we;
  logic ctrl_msg_auth_ready_qs;
  logic ctrl_msg_auth_ready_wd;
  logic ctrl_msg_auth_ready_we;
  logic stat_key_ready_qs;
  logic stat_bdi_ready_qs;
  logic stat_bdo_valid_qs;
  logic stat_end_of_block_qs;
  logic stat_msg_auth_qs;
  logic stat_msg_auth_valid_qs;
  logic [31:0] bdi_qs;
  logic [31:0] bdi_wd;
  logic bdi_we;
  logic [31:0] bdo_qs;
  logic [31:0] bdo_wd;
  logic bdo_we;

  // Register instances

  // Subregister 0 of Multireg pcr
  // R[pcr0]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr0 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[0].de),
    .d      (hw2reg.pcr[0].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[0].q ),

    // to register interface (read)
    .qs     (pcr0_qs)
  );

  // Subregister 1 of Multireg pcr
  // R[pcr1]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr1 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[1].de),
    .d      (hw2reg.pcr[1].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[1].q ),

    // to register interface (read)
    .qs     (pcr1_qs)
  );

  // Subregister 2 of Multireg pcr
  // R[pcr2]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr2 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[2].de),
    .d      (hw2reg.pcr[2].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[2].q ),

    // to register interface (read)
    .qs     (pcr2_qs)
  );

  // Subregister 3 of Multireg pcr
  // R[pcr3]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr3 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[3].de),
    .d      (hw2reg.pcr[3].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[3].q ),

    // to register interface (read)
    .qs     (pcr3_qs)
  );

  // Subregister 4 of Multireg pcr
  // R[pcr4]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr4 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[4].de),
    .d      (hw2reg.pcr[4].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[4].q ),

    // to register interface (read)
    .qs     (pcr4_qs)
  );

  // Subregister 5 of Multireg pcr
  // R[pcr5]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr5 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[5].de),
    .d      (hw2reg.pcr[5].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[5].q ),

    // to register interface (read)
    .qs     (pcr5_qs)
  );

  // Subregister 6 of Multireg pcr
  // R[pcr6]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr6 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[6].de),
    .d      (hw2reg.pcr[6].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[6].q ),

    // to register interface (read)
    .qs     (pcr6_qs)
  );

  // Subregister 7 of Multireg pcr
  // R[pcr7]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RO"),
    .RESVAL  (32'h0)
  ) u_pcr7 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.pcr[7].de),
    .d      (hw2reg.pcr[7].d ),

    // to internal hardware
    .qe     (),
    .q      (reg2hw.pcr[7].q ),

    // to register interface (read)
    .qs     (pcr7_qs)
  );



  // Subregister 0 of Multireg key
  // R[key0]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RW"),
    .RESVAL  (32'h0)
  ) u_key0 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (key0_we),
    .wd     (key0_wd),

    // from internal hardware
    .de     (hw2reg.key[0].de),
    .d      (hw2reg.key[0].d ),

    // to internal hardware
    .qe     (reg2hw.key[0].qe),
    .q      (reg2hw.key[0].q ),

    // to register interface (read)
    .qs     (key0_qs)
  );

  // Subregister 1 of Multireg key
  // R[key1]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RW"),
    .RESVAL  (32'h0)
  ) u_key1 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (key1_we),
    .wd     (key1_wd),

    // from internal hardware
    .de     (hw2reg.key[1].de),
    .d      (hw2reg.key[1].d ),

    // to internal hardware
    .qe     (reg2hw.key[1].qe),
    .q      (reg2hw.key[1].q ),

    // to register interface (read)
    .qs     (key1_qs)
  );

  // Subregister 2 of Multireg key
  // R[key2]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RW"),
    .RESVAL  (32'h0)
  ) u_key2 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (key2_we),
    .wd     (key2_wd),

    // from internal hardware
    .de     (hw2reg.key[2].de),
    .d      (hw2reg.key[2].d ),

    // to internal hardware
    .qe     (reg2hw.key[2].qe),
    .q      (reg2hw.key[2].q ),

    // to register interface (read)
    .qs     (key2_qs)
  );

  // Subregister 3 of Multireg key
  // R[key3]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RW"),
    .RESVAL  (32'h0)
  ) u_key3 (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (key3_we),
    .wd     (key3_wd),

    // from internal hardware
    .de     (hw2reg.key[3].de),
    .d      (hw2reg.key[3].d ),

    // to internal hardware
    .qe     (reg2hw.key[3].qe),
    .q      (reg2hw.key[3].q ),

    // to register interface (read)
    .qs     (key3_qs)
  );


  // R[ctrl]: V(False)

  //   F[key_valid]: 0:0
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_key_valid (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_key_valid_we),
    .wd     (ctrl_key_valid_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.key_valid.de),
    .d      (hw2reg.ctrl.key_valid.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.key_valid.qe),
    .q      (reg2hw.ctrl.key_valid.q ),

    // to register interface (read)
    .qs     (ctrl_key_valid_qs)
  );


  //   F[bdi_valid]: 1:1
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_bdi_valid (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_bdi_valid_we),
    .wd     (ctrl_bdi_valid_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.bdi_valid.de),
    .d      (hw2reg.ctrl.bdi_valid.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.bdi_valid.qe),
    .q      (reg2hw.ctrl.bdi_valid.q ),

    // to register interface (read)
    .qs     (ctrl_bdi_valid_qs)
  );


  //   F[bdi_type]: 5:2
  prim_subreg #(
    .DW      (4),
    .SWACCESS("RW"),
    .RESVAL  (4'h0)
  ) u_ctrl_bdi_type (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_bdi_type_we),
    .wd     (ctrl_bdi_type_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.bdi_type.de),
    .d      (hw2reg.ctrl.bdi_type.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.bdi_type.qe),
    .q      (reg2hw.ctrl.bdi_type.q ),

    // to register interface (read)
    .qs     (ctrl_bdi_type_qs)
  );


  //   F[bdi_size]: 8:6
  prim_subreg #(
    .DW      (3),
    .SWACCESS("RW"),
    .RESVAL  (3'h0)
  ) u_ctrl_bdi_size (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_bdi_size_we),
    .wd     (ctrl_bdi_size_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.bdi_size.de),
    .d      (hw2reg.ctrl.bdi_size.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.bdi_size.qe),
    .q      (reg2hw.ctrl.bdi_size.q ),

    // to register interface (read)
    .qs     (ctrl_bdi_size_qs)
  );


  //   F[bdo_ready]: 9:9
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_bdo_ready (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_bdo_ready_we),
    .wd     (ctrl_bdo_ready_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.bdo_ready.de),
    .d      (hw2reg.ctrl.bdo_ready.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.bdo_ready.qe),
    .q      (reg2hw.ctrl.bdo_ready.q ),

    // to register interface (read)
    .qs     (ctrl_bdo_ready_qs)
  );


  //   F[rst_sync]: 10:10
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_rst_sync (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_rst_sync_we),
    .wd     (ctrl_rst_sync_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.rst_sync.de),
    .d      (hw2reg.ctrl.rst_sync.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.rst_sync.qe),
    .q      (reg2hw.ctrl.rst_sync.q ),

    // to register interface (read)
    .qs     (ctrl_rst_sync_qs)
  );


  //   F[hash_in]: 11:11
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_hash_in (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_hash_in_we),
    .wd     (ctrl_hash_in_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.hash_in.de),
    .d      (hw2reg.ctrl.hash_in.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.hash_in.qe),
    .q      (reg2hw.ctrl.hash_in.q ),

    // to register interface (read)
    .qs     (ctrl_hash_in_qs)
  );


  //   F[decrypt_in]: 12:12
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_decrypt_in (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_decrypt_in_we),
    .wd     (ctrl_decrypt_in_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.decrypt_in.de),
    .d      (hw2reg.ctrl.decrypt_in.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.decrypt_in.qe),
    .q      (reg2hw.ctrl.decrypt_in.q ),

    // to register interface (read)
    .qs     (ctrl_decrypt_in_qs)
  );


  //   F[pcr_update]: 13:13
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_pcr_update (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_pcr_update_we),
    .wd     (ctrl_pcr_update_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.pcr_update.de),
    .d      (hw2reg.ctrl.pcr_update.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.pcr_update.qe),
    .q      (reg2hw.ctrl.pcr_update.q ),

    // to register interface (read)
    .qs     (ctrl_pcr_update_qs)
  );


  //   F[bdi_eot]: 14:14
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_bdi_eot (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_bdi_eot_we),
    .wd     (ctrl_bdi_eot_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.bdi_eot.de),
    .d      (hw2reg.ctrl.bdi_eot.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.bdi_eot.qe),
    .q      (reg2hw.ctrl.bdi_eot.q ),

    // to register interface (read)
    .qs     (ctrl_bdi_eot_qs)
  );


  //   F[bdi_eoi]: 15:15
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_bdi_eoi (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_bdi_eoi_we),
    .wd     (ctrl_bdi_eoi_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.bdi_eoi.de),
    .d      (hw2reg.ctrl.bdi_eoi.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.bdi_eoi.qe),
    .q      (reg2hw.ctrl.bdi_eoi.q ),

    // to register interface (read)
    .qs     (ctrl_bdi_eoi_qs)
  );


  //   F[opcode]: 19:16
  prim_subreg #(
    .DW      (4),
    .SWACCESS("RW"),
    .RESVAL  (4'h0)
  ) u_ctrl_opcode (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_opcode_we),
    .wd     (ctrl_opcode_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.opcode.de),
    .d      (hw2reg.ctrl.opcode.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.opcode.qe),
    .q      (reg2hw.ctrl.opcode.q ),

    // to register interface (read)
    .qs     (ctrl_opcode_qs)
  );


  //   F[msg_auth_ready]: 20:20
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RW"),
    .RESVAL  (1'h0)
  ) u_ctrl_msg_auth_ready (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (ctrl_msg_auth_ready_we),
    .wd     (ctrl_msg_auth_ready_wd),

    // from internal hardware
    .de     (hw2reg.ctrl.msg_auth_ready.de),
    .d      (hw2reg.ctrl.msg_auth_ready.d ),

    // to internal hardware
    .qe     (reg2hw.ctrl.msg_auth_ready.qe),
    .q      (reg2hw.ctrl.msg_auth_ready.q ),

    // to register interface (read)
    .qs     (ctrl_msg_auth_ready_qs)
  );


  // R[stat]: V(False)

  //   F[key_ready]: 0:0
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RO"),
    .RESVAL  (1'h0)
  ) u_stat_key_ready (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.stat.key_ready.de),
    .d      (hw2reg.stat.key_ready.d ),

    // to internal hardware
    .qe     (reg2hw.stat.key_ready.qe),
    .q      (reg2hw.stat.key_ready.q ),

    // to register interface (read)
    .qs     (stat_key_ready_qs)
  );


  //   F[bdi_ready]: 1:1
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RO"),
    .RESVAL  (1'h0)
  ) u_stat_bdi_ready (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.stat.bdi_ready.de),
    .d      (hw2reg.stat.bdi_ready.d ),

    // to internal hardware
    .qe     (reg2hw.stat.bdi_ready.qe),
    .q      (reg2hw.stat.bdi_ready.q ),

    // to register interface (read)
    .qs     (stat_bdi_ready_qs)
  );


  //   F[bdo_valid]: 2:2
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RO"),
    .RESVAL  (1'h0)
  ) u_stat_bdo_valid (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.stat.bdo_valid.de),
    .d      (hw2reg.stat.bdo_valid.d ),

    // to internal hardware
    .qe     (reg2hw.stat.bdo_valid.qe),
    .q      (reg2hw.stat.bdo_valid.q ),

    // to register interface (read)
    .qs     (stat_bdo_valid_qs)
  );


  //   F[end_of_block]: 3:3
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RO"),
    .RESVAL  (1'h0)
  ) u_stat_end_of_block (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.stat.end_of_block.de),
    .d      (hw2reg.stat.end_of_block.d ),

    // to internal hardware
    .qe     (reg2hw.stat.end_of_block.qe),
    .q      (reg2hw.stat.end_of_block.q ),

    // to register interface (read)
    .qs     (stat_end_of_block_qs)
  );


  //   F[msg_auth]: 4:4
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RO"),
    .RESVAL  (1'h0)
  ) u_stat_msg_auth (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.stat.msg_auth.de),
    .d      (hw2reg.stat.msg_auth.d ),

    // to internal hardware
    .qe     (reg2hw.stat.msg_auth.qe),
    .q      (reg2hw.stat.msg_auth.q ),

    // to register interface (read)
    .qs     (stat_msg_auth_qs)
  );


  //   F[msg_auth_valid]: 5:5
  prim_subreg #(
    .DW      (1),
    .SWACCESS("RO"),
    .RESVAL  (1'h0)
  ) u_stat_msg_auth_valid (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    .we     (1'b0),
    .wd     ('0  ),

    // from internal hardware
    .de     (hw2reg.stat.msg_auth_valid.de),
    .d      (hw2reg.stat.msg_auth_valid.d ),

    // to internal hardware
    .qe     (reg2hw.stat.msg_auth_valid.qe),
    .q      (reg2hw.stat.msg_auth_valid.q ),

    // to register interface (read)
    .qs     (stat_msg_auth_valid_qs)
  );


  // R[bdi]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RW"),
    .RESVAL  (32'h0)
  ) u_bdi (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (bdi_we),
    .wd     (bdi_wd),

    // from internal hardware
    .de     (hw2reg.bdi.de),
    .d      (hw2reg.bdi.d ),

    // to internal hardware
    .qe     (reg2hw.bdi.qe),
    .q      (reg2hw.bdi.q ),

    // to register interface (read)
    .qs     (bdi_qs)
  );


  // R[bdo]: V(False)

  prim_subreg #(
    .DW      (32),
    .SWACCESS("RW"),
    .RESVAL  (32'h0)
  ) u_bdo (
    .clk_i   (clk_i    ),
    .rst_ni  (rst_ni  ),

    // from register interface
    .we     (bdo_we),
    .wd     (bdo_wd),

    // from internal hardware
    .de     (hw2reg.bdo.de),
    .d      (hw2reg.bdo.d ),

    // to internal hardware
    .qe     (reg2hw.bdo.qe),
    .q      (reg2hw.bdo.q ),

    // to register interface (read)
    .qs     (bdo_qs)
  );




  logic [15:0] addr_hit;
  always_comb begin
    addr_hit = '0;
    addr_hit[ 0] = (reg_addr == ROT_PCR0_OFFSET);
    addr_hit[ 1] = (reg_addr == ROT_PCR1_OFFSET);
    addr_hit[ 2] = (reg_addr == ROT_PCR2_OFFSET);
    addr_hit[ 3] = (reg_addr == ROT_PCR3_OFFSET);
    addr_hit[ 4] = (reg_addr == ROT_PCR4_OFFSET);
    addr_hit[ 5] = (reg_addr == ROT_PCR5_OFFSET);
    addr_hit[ 6] = (reg_addr == ROT_PCR6_OFFSET);
    addr_hit[ 7] = (reg_addr == ROT_PCR7_OFFSET);
    addr_hit[ 8] = (reg_addr == ROT_KEY0_OFFSET);
    addr_hit[ 9] = (reg_addr == ROT_KEY1_OFFSET);
    addr_hit[10] = (reg_addr == ROT_KEY2_OFFSET);
    addr_hit[11] = (reg_addr == ROT_KEY3_OFFSET);
    addr_hit[12] = (reg_addr == ROT_CTRL_OFFSET);
    addr_hit[13] = (reg_addr == ROT_STAT_OFFSET);
    addr_hit[14] = (reg_addr == ROT_BDI_OFFSET);
    addr_hit[15] = (reg_addr == ROT_BDO_OFFSET);
  end

  assign addrmiss = (reg_re || reg_we) ? ~|addr_hit : 1'b0 ;

  // Check sub-word write is permitted
  always_comb begin
    wr_err = 1'b0;
    if (addr_hit[ 0] && reg_we && (ROT_PERMIT[ 0] != (ROT_PERMIT[ 0] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 1] && reg_we && (ROT_PERMIT[ 1] != (ROT_PERMIT[ 1] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 2] && reg_we && (ROT_PERMIT[ 2] != (ROT_PERMIT[ 2] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 3] && reg_we && (ROT_PERMIT[ 3] != (ROT_PERMIT[ 3] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 4] && reg_we && (ROT_PERMIT[ 4] != (ROT_PERMIT[ 4] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 5] && reg_we && (ROT_PERMIT[ 5] != (ROT_PERMIT[ 5] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 6] && reg_we && (ROT_PERMIT[ 6] != (ROT_PERMIT[ 6] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 7] && reg_we && (ROT_PERMIT[ 7] != (ROT_PERMIT[ 7] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 8] && reg_we && (ROT_PERMIT[ 8] != (ROT_PERMIT[ 8] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[ 9] && reg_we && (ROT_PERMIT[ 9] != (ROT_PERMIT[ 9] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[10] && reg_we && (ROT_PERMIT[10] != (ROT_PERMIT[10] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[11] && reg_we && (ROT_PERMIT[11] != (ROT_PERMIT[11] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[12] && reg_we && (ROT_PERMIT[12] != (ROT_PERMIT[12] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[13] && reg_we && (ROT_PERMIT[13] != (ROT_PERMIT[13] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[14] && reg_we && (ROT_PERMIT[14] != (ROT_PERMIT[14] & reg_be))) wr_err = 1'b1 ;
    if (addr_hit[15] && reg_we && (ROT_PERMIT[15] != (ROT_PERMIT[15] & reg_be))) wr_err = 1'b1 ;
  end









  assign key0_we = addr_hit[8] & reg_we & ~wr_err;
  assign key0_wd = reg_wdata[31:0];

  assign key1_we = addr_hit[9] & reg_we & ~wr_err;
  assign key1_wd = reg_wdata[31:0];

  assign key2_we = addr_hit[10] & reg_we & ~wr_err;
  assign key2_wd = reg_wdata[31:0];

  assign key3_we = addr_hit[11] & reg_we & ~wr_err;
  assign key3_wd = reg_wdata[31:0];

  assign ctrl_key_valid_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_key_valid_wd = reg_wdata[0];

  assign ctrl_bdi_valid_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_bdi_valid_wd = reg_wdata[1];

  assign ctrl_bdi_type_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_bdi_type_wd = reg_wdata[5:2];

  assign ctrl_bdi_size_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_bdi_size_wd = reg_wdata[8:6];

  assign ctrl_bdo_ready_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_bdo_ready_wd = reg_wdata[9];

  assign ctrl_rst_sync_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_rst_sync_wd = reg_wdata[10];

  assign ctrl_hash_in_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_hash_in_wd = reg_wdata[11];

  assign ctrl_decrypt_in_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_decrypt_in_wd = reg_wdata[12];

  assign ctrl_pcr_update_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_pcr_update_wd = reg_wdata[13];

  assign ctrl_bdi_eot_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_bdi_eot_wd = reg_wdata[14];

  assign ctrl_bdi_eoi_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_bdi_eoi_wd = reg_wdata[15];

  assign ctrl_opcode_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_opcode_wd = reg_wdata[19:16];

  assign ctrl_msg_auth_ready_we = addr_hit[12] & reg_we & ~wr_err;
  assign ctrl_msg_auth_ready_wd = reg_wdata[20];







  assign bdi_we = addr_hit[14] & reg_we & ~wr_err;
  assign bdi_wd = reg_wdata[31:0];

  assign bdo_we = addr_hit[15] & reg_we & ~wr_err;
  assign bdo_wd = reg_wdata[31:0];

  // Read data return
  always_comb begin
    reg_rdata_next = '0;
    unique case (1'b1)
      addr_hit[0]: begin
        reg_rdata_next[31:0] = pcr0_qs;
      end

      addr_hit[1]: begin
        reg_rdata_next[31:0] = pcr1_qs;
      end

      addr_hit[2]: begin
        reg_rdata_next[31:0] = pcr2_qs;
      end

      addr_hit[3]: begin
        reg_rdata_next[31:0] = pcr3_qs;
      end

      addr_hit[4]: begin
        reg_rdata_next[31:0] = pcr4_qs;
      end

      addr_hit[5]: begin
        reg_rdata_next[31:0] = pcr5_qs;
      end

      addr_hit[6]: begin
        reg_rdata_next[31:0] = pcr6_qs;
      end

      addr_hit[7]: begin
        reg_rdata_next[31:0] = pcr7_qs;
      end

      addr_hit[8]: begin
        reg_rdata_next[31:0] = key0_qs;
      end

      addr_hit[9]: begin
        reg_rdata_next[31:0] = key1_qs;
      end

      addr_hit[10]: begin
        reg_rdata_next[31:0] = key2_qs;
      end

      addr_hit[11]: begin
        reg_rdata_next[31:0] = key3_qs;
      end

      addr_hit[12]: begin
        reg_rdata_next[0] = ctrl_key_valid_qs;
        reg_rdata_next[1] = ctrl_bdi_valid_qs;
        reg_rdata_next[5:2] = ctrl_bdi_type_qs;
        reg_rdata_next[8:6] = ctrl_bdi_size_qs;
        reg_rdata_next[9] = ctrl_bdo_ready_qs;
        reg_rdata_next[10] = ctrl_rst_sync_qs;
        reg_rdata_next[11] = ctrl_hash_in_qs;
        reg_rdata_next[12] = ctrl_decrypt_in_qs;
        reg_rdata_next[13] = ctrl_pcr_update_qs;
        reg_rdata_next[14] = ctrl_bdi_eot_qs;
        reg_rdata_next[15] = ctrl_bdi_eoi_qs;
        reg_rdata_next[19:16] = ctrl_opcode_qs;
        reg_rdata_next[20] = ctrl_msg_auth_ready_qs;
      end

      addr_hit[13]: begin
        reg_rdata_next[0] = stat_key_ready_qs;
        reg_rdata_next[1] = stat_bdi_ready_qs;
        reg_rdata_next[2] = stat_bdo_valid_qs;
        reg_rdata_next[3] = stat_end_of_block_qs;
        reg_rdata_next[4] = stat_msg_auth_qs;
        reg_rdata_next[5] = stat_msg_auth_valid_qs;
      end

      addr_hit[14]: begin
        reg_rdata_next[31:0] = bdi_qs;
      end

      addr_hit[15]: begin
        reg_rdata_next[31:0] = bdo_qs;
      end

      default: begin
        reg_rdata_next = '1;
      end
    endcase
  end

  // Assertions for Register Interface
  `ASSERT_PULSE(wePulse, reg_we)
  `ASSERT_PULSE(rePulse, reg_re)

  `ASSERT(reAfterRv, $rose(reg_re || reg_we) |=> tl_o.d_valid)

  `ASSERT(en2addrHit, (reg_we || reg_re) |-> $onehot0(addr_hit))

  // this is formulated as an assumption such that the FPV testbenches do disprove this
  // property by mistake
  `ASSUME(reqParity, tl_reg_h2d.a_valid |-> tl_reg_h2d.a_user.parity_en == 1'b0)

endmodule
